package com.bergerkiller.bukkit.nolagg;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.zip.ZipException;

import org.bukkit.Bukkit;
import org.bukkit.craftbukkit.CraftWorld;

import net.minecraft.server.*;

@SuppressWarnings("rawtypes")
public class BufferedChunkRegionLoader extends ChunkRegionLoader {
	
	public static int saved = 0;
			
	public static void restoreAll() {
		if (!NoLagg.useBufferedChunkLoading) return;
		for (org.bukkit.World world : Bukkit.getServer().getWorlds()) {
			restore(world);
		}
	}
	public static void restore(org.bukkit.World world) {
		restore(((CraftWorld) world).getHandle());
	}
	public static void restore(WorldServer world) {
		restore(world.chunkProviderServer);
	}
	public static void restore(ChunkProviderServer cps) {
		if (!NoLagg.useBufferedChunkLoading) return;
		try {
			Field e = ChunkProviderServer.class.getDeclaredField("e");
			e.setAccessible(true);
			IChunkLoader loader = (IChunkLoader) e.get(cps);
			if (!(loader instanceof BufferedChunkRegionLoader)) return;
			
			BufferedChunkRegionLoader bloader = (BufferedChunkRegionLoader) loader;
									
			synchronized (bloader.lock) {
				//save all chunks
				if (!bloader.pendingsave.isEmpty()) {
					NoLagg.log(Level.INFO, "Saving world " + cps.world.getWorld().getName() + " (" + bloader.pendingsave.size() + " chunks)...");
					while (bloader.c());
				}
			}
						
			//create a new loader and set it
			loader = new ChunkRegionLoader(bloader.sourcefolder);
			e.set(cps, loader);
			FileIOThread.a.a((ChunkRegionLoader) loader);	
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to de-initialize the buffered chunk loader:");
			t.printStackTrace();
			NoLagg.useBufferedChunkLoading = false;
		}
	}
	
	public static void convertAll() {
		if (!NoLagg.useBufferedChunkLoading) return;
		for (org.bukkit.World world : Bukkit.getServer().getWorlds()) {
			convert(world);
		}
	}
	public static void convert(org.bukkit.World world) {
		convert(((CraftWorld) world).getHandle());
	}
	public static void convert(WorldServer world) {
		convert(world.chunkProviderServer);
	}
	public static void convert(ChunkProviderServer cps) {
		if (!NoLagg.useBufferedChunkLoading) return;
		try {
			if (eraseChunkMethod == null) {
				eraseChunkMethod = RegionFile.class.getDeclaredMethod("a", int.class, int.class, int.class);
				eraseChunkMethod.setAccessible(true);
			}
			
			Field e = ChunkProviderServer.class.getDeclaredField("e");
			e.setAccessible(true);
			IChunkLoader loader = (IChunkLoader) e.get(cps);
			if (loader instanceof BufferedChunkRegionLoader) return;
			//get needed fields
			Field a = ChunkRegionLoader.class.getDeclaredField("a");
			Field b = ChunkRegionLoader.class.getDeclaredField("b");
			Field c = ChunkRegionLoader.class.getDeclaredField("c");
			Field d = ChunkRegionLoader.class.getDeclaredField("d");
			a.setAccessible(true);
			b.setAccessible(true);
			c.setAccessible(true);
			d.setAccessible(true);
			
			//get old loader variables
			List vara = (List) a.get(loader);
			Set varb = (Set) b.get(loader);
			Object varc = (Object) c.get(loader);
			File vard = (File) d.get(loader);
							
			//fix old loader's data
			synchronized (varc) {
				a.set(loader, new ArrayList());
				b.set(loader, new HashSet());
			}
			
			//create a new loader and set it's native members
			loader = new BufferedChunkRegionLoader(cps.world, vara, varb, varc, vard);
			//set native fields as well
			a.set(loader, vara);
			b.set(loader, varb);
			c.set(loader, varc);
			d.set(loader, vard);
			
			FileIOThread.a.a((BufferedChunkRegionLoader) loader);
			
			e.set(cps, loader);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to initialize the buffered chunk loader:");
			t.printStackTrace();
			NoLagg.useBufferedChunkLoading = false;
		}
	}
	
	private static Method eraseChunkMethod;

	private final WorldServer worldOwner;
	private List savequeue;
	private Set pendingsave;
	private Object lock;
	private File sourcefolder;

	private BufferedChunkRegionLoader(final WorldServer world, List savequeue, Set pendingsave, Object lock, File sourcefolder) {
		super(sourcefolder);
		this.savequeue = savequeue;
		this.pendingsave = pendingsave;
		this.lock = lock;
		this.sourcefolder = sourcefolder;
		this.worldOwner = world;
	}

	private static Field pendinga;
	private static Field pendingb;
	public static ChunkCoordIntPair getPendingA(Object from) {
		try {
			if (pendinga == null) {
				pendinga = from.getClass().getDeclaredField("a");
				pendinga.setAccessible(true);
			}
			return (ChunkCoordIntPair) pendinga.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}
	public static NBTTagCompound getPendingB(Object from) {
		try {
			if (pendingb == null) {
				pendingb = from.getClass().getDeclaredField("b");
				pendingb.setAccessible(true);
			}
			return (NBTTagCompound) pendingb.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}

	@Override
    public boolean c() {
		if (!NoLagg.useBufferedChunkLoading) return super.c();
        Object pendingchunktosave = null;
        ChunkCoordIntPair coord;
        synchronized (this.lock) {
            if (this.savequeue.isEmpty()) {
                return false;
            }

            pendingchunktosave = this.savequeue.remove(0);
            coord = getPendingA(pendingchunktosave);
            this.pendingsave.remove(coord);
        }
        
        if (pendingchunktosave != null) {
            try {
                DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, coord.x, coord.z);
                NBTCompressedStreamTools.a(getPendingB(pendingchunktosave), (DataOutput) dataoutputstream);
                saved++;
                dataoutputstream.flush();
                dataoutputstream.close();
                BufferedChunk chunk = BufferedChunk.removeUnloaded(this.worldOwner, coord.x, coord.z);
                if (chunk != null) {
        			chunk.setFree();
                }
            } catch (Exception exception) {
                exception.printStackTrace();
            }
        }
        return true;
    }
		
	public void clearChunk(int x, int z) throws Throwable {
		RegionFile file = RegionFileCache.a(this.sourcefolder, x, z);
		try {
			eraseChunkMethod.invoke(file, x & 31, z & 31, 0);
		} catch (InvocationTargetException ex) {
			throw ex.getCause();
		}
	}
	
	/*
	 * Chunk at wrong location: relocating code
	 */
	public void relocate(BufferedChunk chunk, NBTTagCompound compound, int realX, int realZ) throws Throwable {
		//is a chunk present at the old location?
		NoLagg.log(Level.WARNING, "Chunk read at " + chunk.toString() + " contains [" + realX + ", " + realZ + "] instead, relocating");
		//a chunk is present at the correct location, alter the x/z location of the data instead
		NBTTagCompound level = compound.getCompound("Level");
		level.setInt("xPos", chunk.x);
		level.setInt("zPos", chunk.z);
		//rewrite the chunk
		DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, chunk.x, chunk.z);
        NBTCompressedStreamTools.a(compound, (DataOutput) dataoutputstream);
        dataoutputstream.close();
	}
	
	public int toChunk(double loc) {
		return MathHelper.floor(loc) >> 4;
	}
	
	@SuppressWarnings("unchecked")
	public void fixEntities(Chunk chunk) {
		for (List<Entity> entitylist : chunk.entitySlices) {
			Iterator<Entity> iter = entitylist.iterator();
			while (iter.hasNext()) {
				Entity e = iter.next();
				if (toChunk(e.locX) == chunk.x) {
					if (toChunk(e.locZ) == chunk.z) {
						continue;
					}
				}
				NoLagg.log(Level.INFO, "Removed entity '" + e.getLocalizedName() + 
						"' from [" + chunk.world.getWorld().getName() + ", " + (int) e.locX + ", " + (int) e.locY + ", " + (int) e.locZ + 
						"] because it was wrongly stored in chunk [" + chunk.x + ", " + chunk.z + "]");
				iter.remove();
			}
		}
	}
		
	/*
	 * This is called when a chunk has to be loaded from file
	 */
	@Override
	public Chunk a(World world, int i, int j) {
		if (NoLagg.useBufferedChunkLoading) {
			try {				
				//was it unloaded?
				BufferedChunk chunk = BufferedChunk.removeUnloaded(world, i, j);
				if (chunk != null) {
					return chunk;
				}
				
				//try to load this chunk from the region file cache
				DataInputStream datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
				if (datainputstream == null) return null;
				
				chunk = new BufferedChunk(world, i, j);
				NBTTagCompound nbttagcompound = null;

				try {
					nbttagcompound = chunk.getBufferedTag(datainputstream);
				} catch (ZipException ex) {
					NoLagg.log(Level.SEVERE, "Chunk data of " + chunk.toString() + " got corrupted (not in GZIP format) and has been removed.");
					this.clearChunk(i, j);
					return null;
				} catch (Exception ex) {
					NoLagg.log(Level.SEVERE, "An exception occurred while reading chunk " + chunk.toString() + ":");
					ex.printStackTrace();
					NoLagg.log(Level.WARNING, "Native methods will now try to read it...");	
					
					//failure, Use native coding instead
					datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
					if (datainputstream == null) return null;
					nbttagcompound = NBTCompressedStreamTools.a((DataInput) datainputstream);
				}
				
				
				//checks and load
				if (nbttagcompound.hasKey("Level")) {
					NBTTagCompound level = nbttagcompound.getCompound("Level");
					if (level.hasKey("Blocks")) {
						//actual chunk loading from NBT compound
						try {
							chunk.load(this, nbttagcompound, level);
							chunk.h();
							fixEntities(chunk);
							return chunk;
						} catch (WrongChunkCoordinatesException ex) {
							//try to load it again; maybe it got freed now?
							//otherwise returns null so the chunk provider can generate it
							return this.a(world, i, j);
						} catch (Exception ex) {
							NoLagg.log(Level.SEVERE, "An exception occurred while loading NBT data of chunk " + chunk.toString() + ":");
							ex.printStackTrace();
							NoLagg.log(Level.WARNING, "Native methods will now try to load it...");	
							//failure or from saving queue, use native coding instead
							Chunk c = ChunkLoader.a(world, level);
							fixEntities(c);
							c.h();
							return c;
						}
					} else {
						System.out.println("Chunk file at " + i + "," + j + " is missing block data, skipping");
						return null;
					}
				} else {
					System.out.println("Chunk file at " + i + "," + j + " is missing level data, skipping");
					return null;
				}
			} catch (Throwable t) {
				NoLagg.log(Level.SEVERE, "Buffered chunk loader faced an incompatibility issue with the current build:");
				t.printStackTrace();
				NoLagg.log(Level.INFO, "Buffered chunk loading has been disabled until a server reload or restart is performed.");
				restoreAll();
				NoLagg.useBufferedChunkLoading = false;
			}
		}
		return super.a(world, i, j);
	}

}
